import json

from datetime import datetime
from collections import defaultdict
from dateutil.parser import parse as dtparse


class PacketInvalidData(Exception):
    """Raised when we try to build a packet with invalid data"""
    pass


class PacketInvalidType(Exception):
    """Raised when we try to create a packet with an invalid type"""


# Closures, for properly handling properties
# in the metaclass
def _make_get_required(key):
    def get_required(self):
        return self._required_data.get(key)
    return get_required


def _make_set_required(key):
    def set_required(self, value):
        self._required_data[key] = value
    return set_required


def _make_del_required(key):
    def del_required(self):
        self._required_data[key] = None
    return del_required


def _make_get_allowed(key):
    def get_allowed(self):
        return self._allowed_data.get(key)
    return get_allowed


def _make_set_allowed(key):
    def set_allowed(self, value):
        self._allowed_data[key] = value
    return set_allowed


class MetaPacket(type):
    """Metaclass for packets.

    Looks at the _data_keys_allowed and _data_keys_required attributes
    when a subclass is declared, then adds class properties that
    stores the information in two separate dictionaries on the object.
    """
    def __new__(cls, name, base, attrs):
        required_fields = attrs.get('_data_keys_required', [])
        allowed_fields = attrs.get('_data_keys_allowed', [])
        for k in required_fields:
            attrs[k] = property(_make_get_required(k),
                                _make_set_required(k),
                                _make_del_required(k))
        for k in allowed_fields:
            attrs[k] = property(_make_get_allowed(k), _make_set_allowed(k))

        # fix expected_replies to add a default timeouts
        expected_replies = attrs.get('_expected_reply', [])
        expected_with_timeouts = []
        for r in expected_replies:
            if isinstance(r, dict):
                expected_with_timeouts.append(r)
            elif isinstance(r, Packet):
                expected_with_timeouts.append(
                    {'type': r._packet_type,
                     'timeout': 30240}
                )
            elif isinstance(r, str):
                expected_with_timeouts.append(
                    {'type': r,
                     'timeout': 30240}
                )
            else:
                raise Exception("Invalid reply_type")
        attrs['expected_reply'] = expected_with_timeouts
        return type.__new__(cls, name, base, attrs)


class Packet(object, metaclass=MetaPacket):
    """
    Generic AMIE packet base class

    Class parameters:
        _packet_type: the type of the packet (string)
        _expected_reply: expected reply types (list[string] or list[Packet type])
        _data_keys_required: Data keys that are required for this packet type
        _data_keys_not_required_in_reply: Data keys that are required but
                                          whose value can be inferred if
                                          this is a reply packet
        _data_keys_allowed: Data keys that are allowed for this packet type


    Args:
        packet_rec_id (str): The ID for this packet
        date (datetime.Datetime): A datetime object representing this packet's date attribute
        additional_data (dict): Body data that is outsite the AMIE spec.
        in_reply_to (str, int, amieclient.Packet): The packet this packet is in response to. Can take a packet, int, string, or None.
    """
    _data_keys_required = []
    _data_keys_not_required_in_reply = []
    _data_keys_allowed = []
    _expected_replies = []

    def __init__(self, packet_rec_id=None, trans_rec_id=None,
                 packet_id=None, transaction_id=None,
                 date=None,
                 additional_data={}, in_reply_to=None,
                 client_state=None, client_json=None,
                 remote_site_name=None, local_site_name=None,
                 originating_site_name=None, outgoing_flag=None,
                 transaction_state=None, packet_state=None,
                 _original_data=None,
                 **kwargs):
        # Set up empty data dicts
        self._required_data = dict()
        self._allowed_data = dict()

        self.packet_rec_id = int(packet_rec_id) if packet_rec_id is not None else None
        self.packet_id = int(packet_id) if packet_id is not None else None
        self.trans_rec_id = int(trans_rec_id) if trans_rec_id is not None else None
        self.transaction_id = int(transaction_id) if transaction_id is not None else None
        self.local_site_name = str(local_site_name) if local_site_name is not None else None
        self.remote_site_name = str(remote_site_name) if remote_site_name is not None else None
        self.originating_site_name = str(originating_site_name) if originating_site_name is not None else None
        self.outgoing_flag = outgoing_flag if outgoing_flag is not None else None
        # TODO make sure these states are valid amie states
        self.transaction_state = str(transaction_state) if transaction_state is not None else None
        self.client_state = str(client_state) if client_state is not None else None
        self.packet_state = str(packet_state) if packet_state is not None else None

        # This one is a property with special...properties. See client_json() 
        # below for details.
        self.client_json = client_json

        # Optionally, save the origininal version of the packet, usually
        # meaning the JSON we got from the server
        self._original_data = _original_data

        self.additional_data = additional_data
        if date is not None:
            self.date = dtparse(date)
        else:
            self.date = None

        if in_reply_to is None or isinstance(in_reply_to, int):
            # If we're given a string, or None, just use that.
            self.in_reply_to_id = in_reply_to
        elif isinstance(in_reply_to, str):
            # If it's a string, make it an int
            self.in_reply_to_id = int(in_reply_to)
        elif hasattr(in_reply_to, 'packet_rec_id'):
            # If we're given a packet object, get the ID
            self.in_reply_to_id = in_reply_to.packet_rec_id
        elif in_reply_to.get('header', {}).get('packet_rec_id'):
            # If we're given a dict-like object, get the ID from the header
            self.in_reply_to_id = in_reply_to['header']['packet_rec_id']

        for key, value in kwargs.items():
            if key in self._data_keys_required or key in self._data_keys_allowed:
                if 'Date' in key:
                    # TODO check if this is a valid assumption
                    setattr(self, key, dtparse(value))
                else:
                    setattr(self, key, value)
            else:
                self.additional_data[key] = value

    @property
    def client_json(self):
        return self._client_json

    @client_json.setter
    def client_json(self, input_json):
        """
        If we're given a string for client_json, assume it's json and
        parse it
        """
        if isinstance(input_json, str):
            self._client_json = json.loads(input_json)
        else:
            self._client_json = input_json

    @classmethod
    def _find_packet_type(cls, packet_or_packet_type):
        """
        Finds the class for the given packet or packet type
        """
        pkt_cls = None
        if type(packet_or_packet_type) == str:
            # We're given a string, search in
            # subclasses
            for subclass in Packet.__subclasses__():
                if subclass._packet_type == packet_or_packet_type:
                    pkt_cls = subclass
                    break
        elif packet_or_packet_type.__class__ in Packet.__subclasses__():
            # We've been given a packet, just get its class attribute
            pkt_cls = packet_or_packet_type.__class__

        if pkt_cls is None:
            # Raise a NotImplementedError if we can't find a subclass
            error_str = "No packet type matches provided '{}'".format(packet_or_packet_type)
            raise PacketInvalidType(error_str)
        return pkt_cls

    @classmethod
    def from_dict(cls, data):
        """
        Generates an instance of an AMIE packet of this type from provided dictionary

        Args:
            data (dict): Packet data
        """
        # Get the subclass that matches this json input
        pkt_class = cls._find_packet_type(data['type'])

        obj = pkt_class(packet_rec_id=data['header']['packet_rec_id'],
                        trans_rec_id=data['header']['trans_rec_id'],
                        packet_id=data['header']['packet_id'],
                        transaction_id=data['header']['transaction_id'],
                        local_site_name=data['header']['local_site_name'],
                        remote_site_name=data['header']['remote_site_name'],
                        originating_site_name=data['header']['originating_site_name'],
                        outgoing_flag=data['header']['outgoing_flag'],
                        transaction_state=data['header']['transaction_state'],
                        packet_state=data['header']['packet_state'],
                        in_reply_to=data['header'].get('in_reply_to'),
                        client_state=data['header'].get('client_state'),
                        client_json=data['header'].get('client_json'),
                        _original_data=data,
                        **data['body'])

        # Return an instance of the proper subclass
        return obj

    @classmethod
    def from_json(cls, json_string):
        """
        Generates an instance of an AMIE packet of this type from provided JSON.
        Basically just a wrapper around from_dict.

        Args:
            json_string (string): JSON data
        """
        data = json.loads(json_string)
        return cls.from_dict(data)

    def reply_packet(self, packet_rec_id=None, packet_type=None, force=False):
        """
        Returns a packet that the current packet would expect as a response,
        with the in_reply_to attribute set to the current packet's ID.

        Generally, most packets only have one kind of expected reply,
        so you should be fine to use reply_packet with just the desired packet_rec_id

        Args:
            packet_rec_id: The ID of the reply packet, if needed
            packet_type: Optionally, the type of the reply packet
            force: will create a reply packet whether or not packet_type is in _expected_reply

        Example:
            >>> my_npc = received_rpc.reply_packet()
        """

        expected_replies = [r['type'] for r in self.expected_reply]
        if packet_type and force:
            # Just do it
            pkt_class = self._find_packet_type(packet_type)
        elif len(expected_replies) == 0:
            # This is a packet that does not expect a response
            raise PacketInvalidType("Packet type '{}' does not expect a reply"
                                    .format(self._packet_type))
        elif len(expected_replies) > 1 and packet_type is None:
            # We have more than one expected reply, but no spec'd type
            # to disambiguate
            raise PacketInvalidType("Packet type '{}' has more than one"
                                    " expected response. Specify a packet type"
                                    " for the reply".format(self._packet_type))
        elif packet_type is not None and packet_type not in expected_replies:
            raise PacketInvalidType("'{}' is not an expected reply for packet type '{}'"
                                    .format(packet_type, self._packet_type))
        else:
            # We have one packet type, or a specified packet type, and it is valid
            if packet_type is None:
                packet_type = expected_replies[0]
            pkt_class = self._find_packet_type(packet_type)
        return pkt_class(packet_rec_id=packet_rec_id, in_reply_to=self.packet_rec_id)

    def as_dict(self):
        """
        This packet, as a dictionary.
        """
        data_body = {}
        # Filter out non-defined items from our data collections, converting
        # if neccessary
        for d in [self._required_data, self._allowed_data, self.additional_data]:
            for k, v in d.items():
                if type(v) == datetime:
                    data_body[k] = v.isoformat()
                elif v is not None:
                    data_body[k] = v

        header = {
            'packet_rec_id': self.packet_rec_id,
            'packet_id': self.packet_id,
            'transaction_id': self.packet_rec_id,
            'trans_rec_id': self.trans_rec_id,
            'expected_reply_list': self._expected_reply,
            'local_site_name': self.local_site_name,
            'remote_site_name': self.remote_site_name,
            'originating_site_name': self.originating_site_name,
            'outgoing_flag': self.outgoing_flag,
            'transaction_state': self.transaction_state,
            'packet_state': self.packet_state,
        }
        if self.date is not None:
            header['date'] = self.date.isoformat()
        if self.in_reply_to_id:
            header['in_reply_to'] = self.in_reply_to_id
        if self.client_state:
            header['client_state'] = self.client_state
        if self.client_json:
            header['client_json'] = self.client_json
        data_dict = {
            'DATA_TYPE': 'packet',
            'type': self.packet_type,
            'body': data_body,
            'header': header
        }

        return data_dict

    def missing_attributes(self):
        """
        Returns a list of attributes that need to be filled out by the user in
        order for this packet to be valid.
        """
        if self.in_reply_to_id:
            reqd = list(set(self._data_keys_required)
                        - set(self._data_keys_not_required_in_reply))
        else:
            reqd = self._data_keys_required

        missing = [r for r in reqd if self._required_data.get(r) is None]
        return missing

    def json(self, **json_kwargs):
        """
        The JSON representation of this AMIE packet
        """
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))

    def validate_data(self, raise_on_invalid=False):
        """
        By default, checks to see that all required data items have a
        defined value, unless in_reply_to is not None (in which case,
        we assume the missing data will be filled in based on the referenced
        packet ID.

        Some packet types will override this function, or add additional checks.
        """
        for k, v in self._required_data.items():
            # If this is a packet in reply to another, and this key is one that
            # the server can infer, skip it.
            if (self.in_reply_to_id and
               k in self._data_keys_not_required_in_reply):
                continue
            # Otherwise, throw an error or return false.
            elif v is None:
                if raise_on_invalid:
                    raise PacketInvalidData('Missing required data field: "{}"'.format(k))
                else:
                    return False
        return True

    @property
    def packet_type(self):
        """
        The AMIE name for this packet type.
        """
        return self._packet_type



import json
from .message import UsageMessage, UsageMessageError
from .record import UsageRecordError


class UsageResponseError(Exception):
    pass


class UsageResponse:
    def __init__(self, message, failed_records):
        self.message = message
        self.failed_records = failed_records

    @classmethod
    def from_dict(cls, input_dict):
        records = [UsageRecordError.from_dict(d) for d in
                   input_dict.get('ValidationFailedRecords', [])]
        message = input_dict['Message']
        return cls(message=message, failed_records=records)

    @classmethod
    def from_json(cls, input_json):
        d = json.loads(input_json)
        return cls.from_dict(d)

    def as_dict(self):
        d = {
            'Message': self.message,
            'ValidationFailedRecords': [r.as_dict() for r in self.failed_records]
        }
        return d

    def json(self, **json_kwargs):
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))

    def __repr__(self):
        return "<UsageResponse: {s.message}>".format(s=self)


class FailedUsageResponse:
    def __init__(self, failed_records):
        self.failed_records = failed_records

    @classmethod
    def from_dict(cls, input_dict):
        records = [UsageRecordError.from_dict(d) for d in
                   input_dict.get('ValidationFailedRecords', [])]
        return cls(failed_records=records)

    @classmethod
    def from_json(cls, input_json):
        d = json.loads(input_json)
        return cls.from_dict(d)

    def as_dict(self):
        d = {
            'FailedRecords': [r.as_dict() for r in self.failed_records]
        }
        return d

    def json(self, **json_kwargs):
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))

    def __repr__(self):
        return "<FailedUsageResponse: {r} records>".format(r=len(self.failed_records))


class UsageStatusResource:
    def __init__(self, resource, loaded_record_count, failed_job_count,
                 total_charge, errors=[]):
        self.resource = resource
        self.loaded_record_count = loaded_record_count
        self.failed_job_count = failed_job_count
        self.total_charge = total_charge
        self.errors = errors

    @classmethod
    def from_dict(cls, input_dict):
        errors = [UsageMessageError.from_dict(d) for d in input_dict.get('Errors', [])]
        return cls(
            resource=input_dict['Resource'],
            loaded_record_count=input_dict['LoadedRecordCount'],
            failed_job_count=input_dict['FailedJobCount'],
            total_charge=input_dict['TotalCharge'],
            errors=errors
        )

    @classmethod
    def from_json(cls, input_json):
        d = json.loads(input_json)
        return cls.from_dict(d)

    def as_dict(self):
        return {
            'Resource': self.resource,
            'LoadedRecordCount': self.loaded_record_count,
            'FailedJobCount': self.failed_job_count,
            'TotalCharge': self.total_charge,
            'Errors': [e.as_dict() for e in self.errors]
        }

    def json(self, **json_kwargs):
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))

    def __repr__(self):
        return "<UsageStatusResource: {s.resource}, {n} errors>".format(s=self, n=len(self.errors))


class UsageStatus:
    def __init__(self, resources):
        self.resources = resources

    @classmethod
    def from_list(cls, input_list):
        resources = input_list
        return cls(
            resources=[UsageStatusResource.from_dict(d) for d in resources]
            )

    @classmethod
    def from_json(cls, input_list):
        d = json.loads(input_list)
        return cls.from_list(d)

    def as_list(self):
        return [r.as_dict() for r in self.resources]

    def json(self, **json_kwargs):
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))

    def __repr__(self):
        return "<UsageStatus: {n} resources>".format(n=len(self.resources))



import json
from datetime import datetime

from ..packet.base import Packet


class Transaction(object):
    def __init__(self, transaction_id, state,
                 originating_site, local_site, remote_site,
                 packets=None):
        self.id = transaction_id
        self.state = state
        self.originating_site = originating_site
        self.local_site = local_site
        self.remote_site = remote_site
        if packets:
            self.packets = packets
        else:
            self.packets = []

    @classmethod
    def from_dict(cls, dict_in):
        tx = cls(
                 transaction_id=dict_in['transaction_id'],
                 state=dict_in['state'],
                 originating_site=dict_in['originating_site_name'],
                 local_site=dict_in['local_site_name'],
                 remote_site=dict_in['remote_site_name'],
                 packets=[Packet.from_dict(p) for p in dict_in['DATA']]
        )
        return tx

    @classmethod
    def from_json(cls, json_in):
        tx_in = json.loads(json_in)
        return cls.from_dict(tx_in)

    def as_dict(self):
        data_dict = {
            'DATA_TYPE': 'transaction',
            'transaction_ID': self.id,
            'originating_site_name':  self.originating_site,
            'local_site_name': self.local_site,
            'remote_site_name': self.remote_site,
            'state': self.state,
            'DATA': [pkt.as_dict() for pkt in self.packets]
        }

        return data_dict

    def json(self):
        data_dict = self.as_dict()
        return json.dumps(data_dict)

class TransactionList(object):
    """
    List of transactions.
    """

    def __init__(self, length, limit, offset, total, transactions=None):
        self.length = length
        self.limit = limit
        self.offset = offset
        self.total = total
        if transactions is not None:
            self.transactions = transactions
        else:
            self.transactions = []

    @classmethod
    def from_dict(cls, dict_in):
        tx_list = cls(
            length=dict_in['length'],
            limit=dict_in['limit'],
            offset=dict_in['offset'],
            total=dict_in['total'],
            transactions=[Transaction.from_dict(d) for d in dict_in['DATA']]
        )
        return tx_list

    @classmethod
    def from_json(cls, json_in):
        txlist_in = json.loads(json_in)
        return cls.from_dict(txlist_in)

    def as_dict(self):
        data_dict = {
            'DATA_TYPE': 'transaction_list',
            'length':  self.legth,
            'limit': self.limit,
            'offset': self.offset,
            'total': self.total,
            'DATA': [txn.as_dict() for txn in self.transactions]
        }
        return data_dict

    def json(self, **json_kwargs):
        """
        The JSON representation of this AMIE transaction
        """
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))



"""
AMIE packets relating to users
"""

from .base import Packet, PacketInvalidData


class NotifyUserModify(Packet):
    _packet_type = 'notify_account_inactivate'
    _expected_reply = [{'type': 'inform_transaction_complete', 'timeout': 30240}]
    _data_keys_required = [
        'ActionType',
        'PersonID',
    ]
    _data_keys_not_required_in_reply = []
    _data_keys_allowed = [
        'AcademicDegree',
        'BusinessPhoneComment',
        'BusinessPhoneExtension',
        'BusinessPhoneNumber',
        'City',
        'Country',
        'Department',
        'DnList',
        'Email',
        'Fax',
        'FirstName',
        'HomePhoneComment',
        'HomePhoneExtension',
        'HomePhoneNumber',
        'LastName',
        'MiddleName',
        'Organization',
        'OrgCode',
        'State'
    ]

    def validate_data(self, raise_on_invalid=False):
        action_type = self.ActionType
        if action_type not in ['add', 'delete', 'replace']:
            if raise_on_invalid:
                error_str = 'Invalid action type for notify_user_modify: "{}"'.format(action_type)
                raise PacketInvalidData(error_str)
            else:
                return False
        return super().validate_data(raise_on_invalid)


class RequestUserModify(Packet):
    _packet_type = 'request_user_modify'
    _expected_reply = [{'type': 'inform_transaction_complete', 'timeout': 30240}]
    _data_keys_required = [
        'ActionType',
        'PersonID',
    ]
    _data_keys_not_required_in_reply = []
    _data_keys_allowed = [
        'AcademicDegree',
        'BusinessPhoneComment',
        'BusinessPhoneExtension',
        'BusinessPhoneNumber',
        'CitizenshipList',
        'CitizenshipList',
        'City',
        'Country',
        'Department',
        'DnList',
        'Email',
        'Fax',
        'FirstName',
        'HomePhoneComment',
        'HomePhoneExtension',
        'HomePhoneNumber',
        'LastName',
        'MiddleName',
        'NsfStatusCode',
        'Organization',
        'OrgCode',
        'State',
        'StreetAddress',
        'StreetAddress2',
        'Title',
        'Zip',
    ]

    def validate_data(self, raise_on_invalid=False):
        action_type = self.ActionType
        if action_type not in ['add', 'delete', 'replace']:
            if raise_on_invalid:
                error_str = 'Invalid action type for request_user_modify: "{}"'.format(action_type)
                raise PacketInvalidData(error_str)
            else:
                return False
        return super().validate_data(raise_on_invalid)



import pytest

from ..packet import (Packet, RequestAccountCreate, Packet, PacketInvalidData,
                      NotifyAccountCreate, NotifyPersonDuplicate,
                      NotifyUserModify, RequestUserModify)
from .fixtures import DEMO_JSON_PKT_1, DEMO_JSON_PKT_2


class TestClient:
    """
    Test packet creation and processing.
    """

    def test_creation(self):
        """
        Test that we create a packet of the proper kind
        """
        packet = Packet.from_dict(DEMO_JSON_PKT_1)
        assert packet.packet_type == 'request_account_create'
        assert isinstance(packet, RequestAccountCreate)

    def test_additional_data(self):
        """
        Test that additional data is stored properly
        DEMO_JSON_PKT_1 contains a parameter that is
        not part of the current AMIE spec, UserFavoriteColor,
        so we check for that
        """
        packet = Packet.from_dict(DEMO_JSON_PKT_1)
        assert 'UserFavoriteColor' in packet.additional_data
        assert packet.additional_data['UserFavoriteColor'] == 'blue'

    def test_required_data_storage(self):
        """
        Required data is stored in the proper location on a created packet
        """
        packet = Packet.from_dict(DEMO_JSON_PKT_1)
        for k in RequestAccountCreate._data_keys_required:
            assert k in packet._required_data
            assert getattr(packet, k) == DEMO_JSON_PKT_1['body'].get(k)
            assert packet._required_data.get(k) == DEMO_JSON_PKT_1['body'].get(k)

    def test_required_data_delete_is_none(self):
        """
        If we 'del' a field of required data from the packet,
        it's replaced with None, not removed from the object
        """
        packet = Packet.from_dict(DEMO_JSON_PKT_1)
        for k in RequestAccountCreate._data_keys_required:
            delattr(packet, k)
            assert getattr(packet, k) is None
            assert packet._required_data[k] is None

    def test_allowed_data_storage(self):
        """
        Parameters that are defined in the AMIE spec but optional are stored
        in the right location.
        """
        packet = Packet.from_dict(DEMO_JSON_PKT_1)
        for k in RequestAccountCreate._data_keys_allowed:
            if k in packet._allowed_data:
                assert getattr(packet, k) == DEMO_JSON_PKT_1['body'].get(k)
                assert packet._allowed_data.get(k) == DEMO_JSON_PKT_1['body'].get(k)

    def test_reply_packet_validation(self):
        """
        The in_reply_to field and the packet type are properly set on a packet
        generated via the reply_packet method. Additionally, that packet passes
        validation check even if required data is missing.
        """
        parent_packet = Packet.from_dict(DEMO_JSON_PKT_1)
        reply_packet = parent_packet.reply_packet()
        assert reply_packet.in_reply_to_id == parent_packet.packet_rec_id
        assert reply_packet.packet_type == 'notify_account_create'
        assert isinstance(reply_packet, NotifyAccountCreate)
        for k in NotifyAccountCreate._data_keys_not_required_in_reply:
            v = getattr(reply_packet, k)
            delattr(reply_packet, k)
            assert reply_packet.validate_data()
            # Replace the data so we're not double-testing
            setattr(reply_packet, k, v)

    def test_reply_packet_missing(self):
        """
        The in_reply_to field and the packet type are properly set on a packet
        generated via the reply_packet method. Additionally, that packet passes
        validation check even if required data is missing.
        """
        parent_packet = Packet.from_dict(DEMO_JSON_PKT_1)
        reply_packet = parent_packet.reply_packet()
        assert reply_packet.in_reply_to_id == parent_packet.packet_rec_id
        assert reply_packet.packet_type == 'notify_account_create'
        assert isinstance(reply_packet, NotifyAccountCreate)
        actually_required_keys = list(set(NotifyAccountCreate._data_keys_required) -
                                      set(NotifyAccountCreate._data_keys_not_required_in_reply))
        for k in actually_required_keys:
            v = getattr(reply_packet, k)
            delattr(reply_packet, k)
            assert k in reply_packet.missing_attributes()
            # Replace the data so we're not double-testing
            setattr(reply_packet, k, v)


    def test_validation_required_data(self):
        """
        Packet validation in the general case fails if required data is missing
        """
        packet = Packet.from_dict(DEMO_JSON_PKT_1)
        for k in RequestAccountCreate._data_keys_required:
            v = getattr(packet, k)
            delattr(packet, k)
            with pytest.raises(PacketInvalidData):
                packet.validate_data(raise_on_invalid=True)
            assert not packet.validate_data()
            # Replace the data so we're not double-testing
            setattr(packet, k, v)

    def test_validation_notify_person_duplicate(self):
        """
        Packet validation for NotifyPersonDuplicate packets
        """
        npd_packet = NotifyPersonDuplicate('12345',
                                           PersonID1='abcde',
                                           PersonID2='abcde',
                                           GlobalID1='abcde',
                                           GlobalID2='abcde'
                                           )

        # Test when everything is provided
        assert npd_packet.validate_data()

        err_regex = r'Must provide either GlobalID[12] or PersonID[12]'
        for pid, gid in [('PersonID1', 'GlobalID1'),
                         ('PersonID2', 'GlobalID2')]:
            # Test when both PID and GID are not provided
            setattr(npd_packet, pid, None)
            setattr(npd_packet, gid, None)
            with pytest.raises(PacketInvalidData, match=err_regex):
                npd_packet.validate_data(raise_on_invalid=True)
            assert not npd_packet.validate_data()

            # Test when only PID is provided
            setattr(npd_packet, pid, 'abcde')
            assert npd_packet.validate_data()
            # Test when only GID is provided
            setattr(npd_packet, gid, 'abcde')
            setattr(npd_packet, pid, None)
            assert npd_packet.validate_data()

    def test_validation_notify_user_modify(self):
        """
        Packet validation for NotifyUserModify packets
        """
        num_packet = NotifyUserModify('12345', PersonID='abcde')
        # test missing action type
        with pytest.raises(PacketInvalidData):
            num_packet.validate_data(raise_on_invalid=True)
        assert not num_packet.validate_data()

        # Test valid action types
        for action_type in ['add', 'delete', 'replace']:
            num_packet.ActionType = action_type
            assert num_packet.validate_data()

        # Test invalid action type
        num_packet.ActionType = 'make_just_a_little_bit_taller'
        with pytest.raises(PacketInvalidData):
            num_packet.validate_data(raise_on_invalid=True)
        assert not num_packet.validate_data()

    def test_validation_request_user_modify(self):
        """
        Packet validation for RequestUserModify packets
        """
        rum_packet = RequestUserModify('12345', PersonID='abcde')
        # test missing action type
        with pytest.raises(PacketInvalidData):
            rum_packet.validate_data(raise_on_invalid=True)
        assert not rum_packet.validate_data()

        # Test valid action types
        for action_type in ['add', 'delete', 'replace']:
            rum_packet.ActionType = action_type
            assert rum_packet.validate_data()

        # Test invalid action type
        rum_packet.ActionType = 'make_just_a_little_bit_taller'
        with pytest.raises(PacketInvalidData):
            rum_packet.validate_data(raise_on_invalid=True)
        assert not rum_packet.validate_data()

    def test_packet_fidelity(self):
        """
        Make sure that the data we put into the packet is the same as the data
        we get out, especially if we have two packets of the same type.
        See https://github.com/XSEDE/amieclient/issues/8 &
        https://github.com/XSEDE/amieclient/commit/7c57ce781b70e5a06a1b3811d1f11d9f99a20893
        for why we check both packets here
        """
        packet_1 = Packet.from_dict(DEMO_JSON_PKT_1)
        packet_2 = Packet.from_dict(DEMO_JSON_PKT_2)

        # Check that the _original_data field matches
        assert packet_1._original_data == DEMO_JSON_PKT_1
        assert packet_2._original_data == DEMO_JSON_PKT_2

        # Check that we process the data properly
        for pkt in [packet_1, packet_2]:
            for k, v in pkt._original_data['header'].items():
                # skip header fields we handle elsewhere
                if k in ['expected_reply_list']:
                    continue
                assert getattr(pkt, k) == v
            for k, v in pkt._original_data['body'].items():
                try:
                    pkt_v = getattr(pkt, k)
                except AttributeError:
                    # Handle additional data that's kept separate
                    pkt_v = pkt.additional_data[k]
                assert pkt_v == v



from .base import Packet


class InformTransactionComplete(Packet):
    _packet_type = 'inform_transaction_complete'
    _expected_reply = [{'type': 'inform_transaction_complete', 'timeout': 30240}]
    _data_keys_required = ['DetailCode', 'Message', 'StatusCode']
    _data_keys_not_required_in_reply = []
    _data_keys_allowed = []



import json

from .base import Packet


class PacketList(object):
    """
    A list of packets.
    """
    def __init__(self, message=None, packets=None):
        self.message = message
        if packets is not None:
            self.packets = packets
        else:
            self.packets = []

    @classmethod
    def from_dict(cls, dict_in):
        pkt_list = cls(
            message=dict_in.get('message', ''),
            packets=[Packet.from_dict(d) for d in dict_in['result']]
        )
        return pkt_list

    @classmethod
    def from_json(cls, json_in):
        pkt_list_in = json.loads(json_in)
        return cls.from_dict(pkt_list_in)

    def as_dict(self):
        data_dict = {
            'message': self.message,
            'result': [pkt.as_dict()for pkt in self.packets]
        }
        return data_dict

    def json(self, **json_kwargs):
        """
        The JSON representation of these AMIE packets
        """
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='amieclient',
    version='0.3.2',
    packages=find_packages(),
    install_requires=[
        'requests>=2.20.0,<3',
        'python-dateutil>=2.6.1,<2.7'
    ],
    author='G. Ryan Sablosky',
    author_email='sablosky@psc.edu',
    python_requires='>=3.5',
    description='Library for the XSEDE AMIE REST API.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='Apache Software License v2.0',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    project_urls={
        'Documentation & Examples': 'https://xsede.github.io/amieclient/',
        'Source': 'https://github.com/xsede/amieclient/',
        'Tracker': 'https://github.com/xsede/amieclient/issues',
    },
)



from ..client import AMIEClient
from ..packet import RequestAccountCreate, PacketList
from .fixtures import DEMO_JSON_PKT_1, DEMO_JSON_PKT_LIST


class TestAMIEClient:

    def test_creation(self):
        client = AMIEClient(site_name='test', api_key='test')
        assert client._session.headers['XA-SITE'] == 'test'
        assert client._session.headers['XA-API-KEY'] == 'test'

    def test_amie_url(self):
        client1 = AMIEClient(site_name='test', api_key='test',
                             amie_url='http://localhost/amie')
        assert client1.amie_url == 'http://localhost/amie/'
        client2 = AMIEClient(site_name='test', api_key='test',
                             amie_url='http://localhost/amie/')
        assert client2.amie_url == 'http://localhost/amie/'

    def test_get_packet(self, requests_mock):
        client = AMIEClient(site_name='test', api_key='test')

        # Make the mock request
        packet_url = 'https://amieclient.xsede.org/v0.10/packets/test/12345'
        requests_mock.get(packet_url, json=DEMO_JSON_PKT_1)

        packet = client.get_packet(packet_rec_id='12345')
        assert packet.packet_type == 'request_account_create'
        assert isinstance(packet, RequestAccountCreate)

    def test_get_packet_list(self, requests_mock):
        client = AMIEClient(site_name='test', api_key='test')

        # Make the mock request
        packet_url = 'https://amieclient.xsede.org/v0.10/packets/test'
        requests_mock.get(packet_url, json=DEMO_JSON_PKT_LIST)

        r = client.list_packets()
        assert isinstance(r, PacketList)
        assert r.packets[0].packet_type == 'request_account_create'
        assert r.packets[1].packet_type == 'request_account_create'



import json

from .record import _type_lookup, UsageRecord, UsageRecordError


class UsageMessageException(Exception):
    """
    Exception for invalid data in a UsageMessage
    """
    pass


class _UsageRecordList:
    def __init__(self, in_list=[]):
        self._list = in_list
        self._record_type = None
        if not self._check_usage_type():
            self._list = []
            raise UsageMessageException('Cannot create a UsageMessage with mixed types')

    def _check_usage_type(self):
        if self._record_type is None and len(self._list) > 0:
            rt = self._list[0].record_type.lower().capitalize()
            if rt not in ['Compute', 'Storage', 'Adjustment']:
                raise UsageMessageException('Invalid usage type {}'
                                            .format(rt))
            self._record_type = rt
        # iterate over list records and check against stored type
        return all([x.record_type.lower().capitalize() == self._record_type for x in self._list])

    def append(self, item):
        if not isinstance(item, UsageRecord):
            raise UsageMessageException("Can't add something that isn't a UsageRecord")
        if (self._record_type is not None and
                item.record_type != self._record_type):
            raise UsageMessageException("Can't add a {} record to a {} message"
                                        .format(item.record_type, self._record_type))
        self._list.append(item)
        if not self._check_usage_type():
            self._list.pop()
            raise UsageMessageException('Cannot create a UsageMessage with mixed types')

    def extend(self, items):
        for item in items:
            self.append(item)

    def __getitem__(self, i):
        return self._list.__getitem__(i)

    def __len__(self):
        return len(self._list)

    def __repr__(self):
        return "<UsageRecordList: {s._record_type} type, length {l}>".format(s=self, l=len(self))


class UsageMessage:
    def __init__(self, records):
        self.records = _UsageRecordList(records)

    @classmethod
    def from_dict(cls, input_dict):
        """
        Returns a UsageMessage from a provided dictionary
        """
        ut = input_dict['Records'][0]['UsageType']
        ur_class = _type_lookup(ut)
        records = [ur_class.from_dict(d) for d in input_dict['Records']]
        return cls(records)

    @classmethod
    def from_json(cls, input_json):
        input_dict = json.loads(input_json)
        return cls.from_dict(input_dict)

    def as_dict(self):
        """
        Returns a dictionary version of this record
        """
        d = {
            'UsageType': self.records._record_type,
            'Records': [r.as_dict() for r in self.records]
        }
        return d

    def json(self, **json_kwargs):
        """
        Returns a json version of this message
        """
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))

    def _chunked(self, chunk_size=1000):
        """
        Generator that yields UsageMessages with a maximum of chunk_size number
        UsageRecords. Useful for not going over the 256kb POST limit
        """
        for i in range(0, len(self.records), chunk_size):
            r = self.records[i:i+chunk_size]
            yield self.__class__(r)


class UsageMessageError:
    def __init__(self, error, message):
        self._error = error
        self.message = message

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, _):
        pass

    @classmethod
    def from_dict(cls, input_dict):
        error = input_dict.pop('Error', None)
        message = UsageMessage.from_dict(input_dict)
        return cls(error, message)

    @classmethod
    def from_json(cls, input_json):
        d = json.loads(input_json)
        return cls.from_dict(d)

    def as_dict(self):
        d = {'Error': self._error}
        d.update(self.message.as_dict())
        return d

    def json(self, **json_kwargs):
        return json.dumps(self.as_dict(), **json_kwargs)

    def pretty_print(self):
        """
        prints() a pretty version of the JSON of this packet
        """
        print(self.json(indent=4, sort_keys=True))

    def __repr__(self):
        return "<UsageMessageError: {s._error}>".format(s=self)
